/* -*- mode: c++; c-basic-offset: 3; -*- */
#include "framecache.hh"
#include "ParseLine.hh"
#include "scandir.hh"
#include <sstream>
#include <iostream>
#include <algorithm>
#include <stdexcept>

using namespace std;

static const char* localfile = "file://localhost/";

//=====================================  frame cache constructor.
framecache::framecache(void) 
   : debugLevel(0)
{}

//=====================================  frame cache constuctor.
framecache::framecache(const std::string& file) 
   : debugLevel(0)
{
   string path = file;
   while (! path.empty()) {
      string::size_type ipos = path.find(',');
      string nx_path = path.substr(0, ipos);
      cout << "parse file: " << nx_path << endl;
      if (scandir::test_directory(nx_path)) parse_directory(nx_path);
      else                                  parse_cacheFile(nx_path);
      if (ipos != string::npos) ipos++;
      path.erase(0, ipos);
   }
   coalesce();
}

//=====================================  Coalesce the cache contents.
void
framecache::add_group(const string& prefix, gps_type start, gps_type stop, 
		       gps_type delta, const string& dir) {
   if (debugLevel) {
      cout << "frame group prefix: " << prefix << " start: " << start
	   << " stop: " << stop << " delta: " << delta << " dir: " << dir
	   << endl;
   }
   frame_group fg(prefix, start, stop, delta, dir);
   int adjid = find_adjacent(fg);
   if (adjid < 0) frameCache.push_back(fg);
   else           frameCache[adjid].combine(fg);
}

//=====================================  Coalesce the cache contents.
void
framecache::coalesce(void) {
   size_t N=frameCache.size();
   if (N < 2) return;

   //----------------------------------  Sort the cache vector
   std::sort(frameCache.begin(), frameCache.end());

   size_t i=0;
   for (size_t j=1; j<N; j++) {
      if (frameCache[i].adjacent(frameCache[j])) {
	 frameCache[i].combine(frameCache[j]);
      } else {
	 i++;
	 if (i < j) frameCache[i] = frameCache[j];
      }
   }

   //---------------------------------  Erase coalesced entries.
   i++;
   if (i != N) frameCache.erase(frameCache.begin()+i, frameCache.begin()+N);
}

//=====================================  Display the cache contents.
void
framecache::display(void) const {
   cout << "prefix start stop dt dir" << endl;
   int N = frameCache.size();
   for (int i=0; i<N; i++) {
      frameCache[i].display();
   }
}

//=====================================  Find an adjacent frame group.
int 
framecache::find_adjacent(const frame_group& grp) const {
   int N = frameCache.size();
   for (int i=0; i<N; i++) {
      if (frameCache[i].adjacent(grp)) return i;
   }
   return -1;
}

//=====================================  frame cache constructor.
void
framecache::parse_directory(const std::string& dir) {
   if (debugLevel) cout << "wframecache: parsing directory: " << dir << endl;
   scandir sd(dir);
   while (sd.next_entry()) {
      if (sd.is_frame()) {
	 gps_type gps = sd.gps();
	 gps_type dt  = sd.duration();
	 add_group(sd.prefix(), gps, gps+dt, dt, dir);
      } else if (sd.is_dir()) {
	 string subdir = sd.file_path();
	 size_t dot = subdir.find_last_of("/");
	 if (dot != string::npos && subdir[dot+1] == '.') continue;
	 parse_directory(subdir);
      }
   }
}

//=====================================  Parse a frame cache file.
void
framecache::parse_cacheFile(const std::string& file) {
   size_t lenlocal=string(localfile).size();
   ParseLine pl(file.c_str());
   if (!pl.isOpen()) {
      cerr << "Unable to open frame cache file: " << file << endl;
      throw runtime_error("Can not open frame cache file");
   }
   while (pl.getLine() >= 0) {
      int nArg = pl.getCount();
      if (!nArg) continue;

      //--------------------------------  Parse ligo_data_find cache files
      if (nArg == 5) {
	 //-----------------------------  Extract the directory path
	 string path=pl[4];
	 if (path.substr(0, lenlocal) != localfile) {
	    cerr << "cache path in " << file << "  is not '" << localfile
		 << "'." << endl;
	    break;
	 }
	 path.erase(0, lenlocal-1);
	 size_t inx = path.find_last_of('/');
	 if (inx == string::npos) inx = 0;

	 //-----------------------------  Calculate the other parameters.
	 string prefix = pl[0];
	 prefix += string("-") + pl[1];
	 gps_type start = pl.getInt(2);
	 gps_type delta = pl.getInt(3);
	 gps_type stop  = start + delta;
	 add_group(prefix, start, stop, delta, path.substr(0,inx));
      }

      //--------------------------------  Parse prefix, start, stop, delta, dir
      else if (nArg == 6) {
	 string prefix = pl[0];
	 prefix += string("-") + pl[1];
	 string path = pl[5];
	 if (path.substr(0, lenlocal) == localfile) path.erase(0, lenlocal-1);
	 add_group(prefix, pl.getInt(2), pl.getInt(3), pl.getInt(4), path);
      }

      else {
	 cerr << "Invalid format line in: " << file << endl;
	 break;
      }
   }
}

//=====================================  get a list of frames with specified 
//                                       data
int
framecache::get_list(const std::string& prefix, gps_type start, 
		      gps_type stop, str_vect& frames) const {
   frames.clear();
   size_t N = frameCache.size();
   for (size_t i=0; i<N ; i++) {
      if (frameCache[i].valid(prefix, start, stop)) {
	 frameCache[i].get_frames(start, stop, frames);
      }
   }
   return frames.size();
}

//======================================  Set the debug level;
void
framecache::set_debug(int lvl) {
   debugLevel = lvl;
}

//=====================================  Write a frame cache file
void
framecache::write(const std::string& path) const {
   ofstream out(path.c_str());
   for (auto p=frameCache.begin(); p != frameCache.end(); p++) {
      p->write(out);
   }	
}

//=====================================  construct a frame group
framecache::frame_group::frame_group(const std::string& prefix,
				      gps_type start, 
				      gps_type stop,
				      gps_type frame_len, 
				      const std::string& dir) 
   : _prefix(prefix), _directory(dir), _start(start), _length(frame_len),
     _end(stop)
{}

//=====================================  test for adjacent frame group
bool 
framecache::frame_group::adjacent(const frame_group& grp) const {
   if (grp._end < _start
       || _end < grp._start 
       || _prefix != grp._prefix
       || _directory != grp._directory
       || _length != grp._length)
      return false;
   return true;
}

//=====================================  test for adjacent frame group
void
framecache::frame_group::combine(const frame_group& grp) {
   if (!adjacent(grp)) {
      throw runtime_error("framecache: can't merge groups");
   }
   if (grp._start < _start) _start = grp._start;
   if (_end < grp._end)     _end   = grp._end;
}

//=====================================  Display the frame group contents.
void
framecache::frame_group::display(void) const {
   cout << _prefix << " " << _start << " " << _end << " " << _length 
	<< " " << _directory << endl;
}

//=====================================  Test if requested data in group
bool 
framecache::frame_group::valid(const string& prefix, gps_type start, 
				gps_type stop) const {

   bool rc = (prefix == _prefix && start < _end && stop > _start);
   return rc;
}

//=====================================  Write out the group file entry
void
framecache::frame_group::write(ostream& out) const {
   size_t dash = _prefix.find("-");
   out << _prefix.substr(0, dash) << " " << _prefix.substr(dash+1) << " "
       << _start << " " << _end << " " << _length << " " << localfile
       << _directory.substr(1) << endl;
}

//=====================================  Get a list of frames
int
framecache::frame_group::get_frames(gps_type start, gps_type stop, 
				     str_vect& frames) const {
   int count = 0;

   //-----------------------------------  Handle special entries.
   if (_length == 0) {
      frames.push_back(_directory);
      count++;
   } 

   //-----------------------------------  Build frame list.
   else {
      for (gps_type gps=_start; gps<_end; gps += _length) {
	 if (gps < stop && gps + _length > start) {
	    frames.push_back(frame(gps));
	    count++;
	 }
      }
   }
   return count;
}

//======================================  Get name os frame
std::string 
framecache::frame_group::frame(gps_type gps) const {
   ostringstream ostr;
   ostr << _directory << "/" << _prefix << "-" << gps << "-" 
	<< _length << ".gwf";
   return ostr.str();
}
